local lfs = require "lfs"
local posix = require "posix"

function removedir(dir)
   for file in lfs.dir(dir) do
	  local file_path = dir .. '/' .. file
	  if file ~= "." and file ~= ".." then
		 if lfs.attributes(file_path, 'mode') == 'file' then
			os.remove(file_path)
		 elseif lfs.attributes(file_path, 'mode') == 'directory' then
			removedir(file_path)
		 end
	  end
   end
   lfs.rmdir(dir)
end
   
describe("tkts CLI", function()
			local test_home = "/tmp/tkts_test"
			local tkts_cmd = "./src/./tkts.fnl"

			describe("init operations", function()
						it("should remove existing config directory & recreate it", function()
							  -- Remove test directory if it exists
							  if lfs.attributes(test_home) then
								 removedir(test_home)
							  end
							  
							  -- Create fresh test directory
							  lfs.mkdir("/tmp/tkts_test")
							  lfs.mkdir("/tmp/tkts_test/.config")
							  posix.setenv("HOME", test_home)
							  
							  local init_output = io.popen(tkts_cmd):read("*a")
							  assert.matches("DB{Migrating database from version (%d+) to (%d+)}", init_output)
							  assert.matches("Open: 0 | Closed: 0", init_output)
							  assert.is_true(lfs.attributes(test_home .. "/.config/tkts/tkts.conf", 'mode') == "file")
							  assert.is_true(lfs.attributes(test_home .. "/.config/tkts/tkts.db", 'mode') == "file")
							  assert.is_true(lfs.attributes(test_home .. "/.config/tkts/invoice.tf", 'mode') == "file")
						end)
			end)

			describe("ticket operations", function()
						it("should create and show a ticket", function()
							  local create_output = io.popen(tkts_cmd .. " create -i 'Test Ticket' -d 'Test Description'"):read("*a")
							  assert.matches("T%d+ {:Status Open}", create_output)

							  local tnum = create_output:match("T(%d+)")
							  assert.is_not_nil(tnum)

							  local info_output = io.popen(tkts_cmd .. " info -t " .. tnum):read("*a")
							  assert.matches("Issue: Test Ticket", info_output)
							  assert.matches("Status: Open", info_output)
							  assert.matches("Desc: .*Test Description", info_output)

							  --TODO: Simulate edit tests using ed

							  local close_output = io.popen(tkts_cmd .. " close -t " .. tnum):read("*a")
							  assert.matches("T(%d+) {:Status Closed}", close_output)

							  local reopen_output = io.popen(tkts_cmd .. " reopen -t " .. tnum):read("*a")
							  assert.matches("T(%d+) {:Status Open}", reopen_output)

							  local log_output = io.popen(tkts_cmd .. " log -t " .. tnum .. " -l 'This is a test log record' --tag 'test'"):read("*a")
							  assert.matches("T(%d+) {:log (%d+)}", log_output)
						end)
			end)

			describe("company operations", function()
						it("should create and show a company", function()
							  local create_output = io.popen(tkts_cmd .. " company:create -n 'Socratic Systems' --abv 'SOCSYS' --addr '123 Republic Way, Ancient Greece'"):read("*a")
							  assert.matches("{:Company Socratic Systems}", create_output)
							  --assert.matches("C:%d+ {:Status Created}", create_output)

							  --FIXME: homogenize the output for this module
							  --local comp_id = create_output:match("C(%d+)")
							  --assert.is_not_nil(comp_id)

							  local info_output = io.popen(tkts_cmd .. " company:info -n 'Socratic Systems'"):read("*a")
							  assert.matches("Name: Socratic Systems", info_output)
							  assert.matches("Abbrv: SOCSYS", info_output)
							  assert.matches("Address: 123 Republic Way, Ancient Greece", info_output)
							  assert.matches("Comp ID: (%d+)", info_output)

							  --TODO: Simulate edit tests using ed
						end)
			end)

			describe("client operations", function()
						it("should create and show a client", function()
							  local create_output = io.popen(tkts_cmd .. " client:create -n 'Glaucon' -c 'Socratic Systems'"):read("*a")
							  assert.matches("{:Client Glaucon}", create_output)
							  --assert.matches("CL:%d+ {:Status Created}", create_output)

							  --FIXME: homogenie the output for this module
							  --local client_id = create_output:match("CL(%d+)")
							  --assert.is_not_nil(client_id)

							  local info_output = io.popen(tkts_cmd .. " client:info -n 'Glaucon' -c 'Socratic Systems'"):read("*a")
							  assert.matches("Name: Glaucon", info_output)
							  assert.matches("Company: Socratic Systems", info_output)
							  assert.matches("Role:", info_output)
							  assert.matches("Phone:", info_output)
							  assert.matches("Email:", info_output)
							  assert.matches("Client ID: (%d+)", info_output)

							  --TODO: Simulate edit tests using ed
						end)
			end)

			describe("project operations", function()
						it("should create two tickets, assign them to a project, and show the project", function()
							  local ticketa_output = io.popen(tkts_cmd .. " create -i 'Ticket A' -d 'Project Ticket A'"):read("*a")
							  assert.matches("T%d+ {:Status Open}", ticketa_output)
							  local ticketb_output = io.popen(tkts_cmd .. " create -i 'Ticket B' -d 'Project Ticket B'"):read("*a")
							  assert.matches("T%d+ {:Status Open}", ticketb_output)

							  local tnuma = ticketa_output:match("T(%d+)")
							  assert.is_not_nil(tnuma)
							  local tnumb = ticketb_output:match("T(%d+)")
							  assert.is_not_nil(tnumb)

							  local create_project = io.popen(tkts_cmd .. " project:create -n 'Tickets of the Republic' -c 'Socratic Systems' --client 'Glaucon' -d 'Explaining the Rhetoric of the Cave'"):read("*a")

							  local pnum = create_project:match("P(%d+)")
							  assert.is_not_nil(pnum)

							  local add_ticketa = io.popen(tkts_cmd .. " project:add-tkt -p " .. pnum .. " -t " .. tnuma):read("*a")
							  assert.matches("P(%d+) {:Status Modified :add T(%d+)}", add_ticketa)
							  local add_ticketb = io.popen(tkts_cmd .. " project:add-tkt -p " .. pnum .. " -t " .. tnumb):read("*a")
							  assert.matches("P(%d+) {:Status Modified :add T(%d+)}", add_ticketb)

							  local project_status = io.popen(tkts_cmd .. " project:status -p " .. pnum):read("*a")
							  assert.matches("P(%d+)%) %[Socratic Systems%] %- Tickets of the Republic", project_status)
							  assert.matches("Status: Active", project_status)
							  assert.matches("Description: Explaining the Rhetoric of the Cave", project_status)
							  assert.matches("Client: Glaucon", project_status)
							  assert.matches("Total Time:.*", project_status)
							  assert.matches("T(%d+): Ticket A %(Open%)", project_status)
							  assert.matches("T(%d+): Ticket B %(Open%)", project_status)

							  local remove_ticketa = io.popen(tkts_cmd .. " project:rm-tkt -p " .. pnum .. " -t " .. tnuma):read("*a")
							  assert.matches("P(%d+) {:Status Modified :rm T(%d+)}", remove_ticketa)

							  local project_status_removed = io.popen(tkts_cmd .. " project:status -p " .. pnum):read("*a")
							  assert.not_matches("T(%d+): Ticket A %(Open%)", project_status_removed)

							  local project_hold = io.popen(tkts_cmd .. " project:hold -p " .. pnum):read("*a")
							  assert.matches("P(%d+) {:Status On Hold}", project_hold)

							  local project_reactivate = io.popen(tkts_cmd .. " project:reactivate -p " .. pnum):read("*a")
							  assert.matches("P(%d+) {:Status Active}", project_reactivate)

							  local project_complete = io.popen(tkts_cmd .. " project:complete -p " .. pnum):read("*a")
							  assert.matches("P(%d+) {:Status Completed}", project_complete)
						end)
			end)

			describe("deinit operations", function()
						it("should remove existing config directory", function()
							  -- Remove test directory if it exists
							  if lfs.attributes(test_home) then
								 removedir(test_home)
							  end
						end)
			end)
end)
