#!/usr/bin/fennel5.4
;;===== Deps =====
(local lume (require :lume))
(local lyaml (require :lyaml))
(local sql (require :lsqlite3))
(local inspect (require :inspect))

;;===== Global Vars =====
(global ver {"tkts" 3.0})

(global dot {"dir" (.. (os.getenv "HOME") "/.config/tkts/")
             "file" (.. (os.getenv "HOME") "/.config/tkts/tkts.conf")})

(var conf {"editor" (or (os.getenv "EDITOR") "ed")
           "mail" "mutt -s"
           "rate" "0.00"
           "user" (or (os.getenv "USER") "")
           "company" "Change Me"
           "tkt_loc" (.. dot.dir "tkts.db")
           "invoice_template" (.. dot.dir "invoice.tf")})

;;===== Invoice Template =====
(global invoice-defaults
        {:tax_rate 0.0
         :payment_terms 30
         :currency "$"
         :template :default})

(global invoice_templates
        {:default ".lj
.ps +3
.UL \"{full_name}\"
\\*<
.br
{address}
.br
{cell_number}
\\*>

.lj
.TS
tab(+) expand;
lUP14 s s s .
Bill To+++
_
.T&
l s lB r .
{company}+Invoice No:+{invoice_number}
{company_addr}+Invoice Date:+{issued_date}
{company_zip}+Due Date:+{due_date}
{client_email}+Period:+{period}
{client_cellphone}++
.TE

.TS
tab(@) expand box;
cB | cB | cB | cB
l | n | n | n .
Description@QTY/HR@Unit Price@Total
_
{line_items}
_
.T&
c c | rB r .
@@Subtotal@${total_cost}
@@Tax Rate@{tax_rate}%
@@Total Tax@${tax_amount}
@@Balance Due@${final_amount}
.TE"})

(global ticket-table-template "
.ps +1
.B \"T{tkt_num} {tkt_title}\"
.ps -1
.sp 0.2i
.TS
tab(@) expand box;
cB | cB | cB | cB
l | n | n | n .
Description@QTY/HR@Unit Price@Total
_
{line_items}
_
.TE")

;; === Schema Versions ===
;; Each schema version represents a complete state of the database
;; Add new versions at the bottom with corresponding schema changes
(local schema-versions
       {1 {"version"
           "CREATE TABLE IF NOT EXISTS version (
         db_ver INTEGER NOT NULL);"
           "tickets"
           "CREATE TABLE IF NOT EXISTS tickets (
         id INTEGER PRIMARY KEY,
         status TEXT CHECK(status IN ('Open', 'Closed')) DEFAULT 'Open' NOT NULL,
         title TEXT NOT NULL,
         desc TEXT,
         client TEXT,
         project TEXT,
         owner TEXT NOT NULL,
         time INTEGER DEFAULT 0 NOT NULL,
         created_on DATETIME DEFAULT CURRENT_TIMESTAMP);"
           "records"
           "CREATE TABLE IF NOT EXISTS records (
         id INTEGER PRIMARY KEY,
         tkt_num INTEGER NOT NULL,
         log TEXT NOT NULL,
         user TEXT NOT NULL,
         time INTEGER DEFAULT 0 NOT NULL,
         created_on DATETIME DEFAULT CURRENT_TIMESTAMP,
         FOREIGN KEY (tkt_num) REFERENCES tickets (id) ON DELETE CASCADE);"
           "company"
           "CREATE TABLE IF NOT EXISTS company (
         id INTEGER PRIMARY KEY,
         name TEXT UNIQUE NOT NULL,
         abbrv TEXT NOT NULL,
         address TEXT,
         created_on DATETIME DEFAULT CURRENT_TIMESTAMP);"
           "client"
           "CREATE TABLE IF NOT EXISTS client (
         id INTEGER PRIMARY KEY,
         name TEXT NOT NULL,
         company TEXT NOT NULL,
         role TEXT,
         phone TEXT,
         email TEXT,
         created_on DATETIME DEFAULT CURRENT_TIMESTAMP,
         FOREIGN KEY (company) REFERENCES company (abbrv) ON DELETE CASCADE);"}

        2 {"tasks"
           "CREATE TABLE IF NOT EXISTS tasks (
         id INTEGER PRIMARY KEY,
         tkt_num INTEGER NOT NULL,
         title TEST NOT NULL,
         state INTEGER DEFAULT 0 NOT NULL,
         created_on DATETIME DEFAULT CURRENT_TIMESTAMP,
         FOREIGN KEY (tkt_num) REFERENCES tickets (id) ON DELETE CASCADE)
         CHECK (state >= 0 AND state <= 2);"}

        3 {"ticket_tags"
           "CREATE TABLE IF NOT EXISTS ticket_tags (
         id INTEGER PRIMARY KEY,
         tkt_num INTEGER NOT NULL,
         tag TEXT NOT NULL,
         FOREIGN KEY (tkt_num) REFERENCES tickets (id) ON DELETE CASCADE);"}

        4 {"record_metadata"
           "ALTER TABLE records ADD COLUMN tag TEXT;"}

        5 {"invoices"
           "CREATE TABLE IF NOT EXISTS invoices (
         id INTEGER PRIMARY KEY,
         tkt_num INTEGER NOT NULL,
         client TEXT NOT NULL,
         start_date TEXT NOT NULL,
         end_date TEXT NOT NULL,
         amount INTEGER NOT NULL,
         content TEXT NOT NULL,
         created_on DATETIME DEFAULT CURRENT_TIMESTAMP,
         modified_on DATETIME DEFAULT CURRENT_TIMESTAMP,
         FOREIGN KEY (tkt_num) REFERENCES tickets (id) ON DELETE CASCADE);"}

        6 {"projects"
           "CREATE TABLE IF NOT EXISTS projects (
        id INTEGER PRIMARY KEY,
        name TEXT NOT NULL,
        description TEXT,
        status TEXT CHECK(status IN ('Active', 'On Hold', 'Completed')) DEFAULT 'Active',
        client TEXT,
        company TEXT,
        start_date TEXT,
        end_date TEXT,
        created_on DATETIME DEFAULT CURRENT_TIMESTAMP,
        FOREIGN KEY (company) REFERENCES company (abbrv));"

           "project_tickets"
           "CREATE TABLE IF NOT EXISTS project_tickets (
        project_id INTEGER NOT NULL,
        ticket_id INTEGER NOT NULL,
        added_on DATETIME DEFAULT CURRENT_TIMESTAMP,
        FOREIGN KEY (project_id) REFERENCES projects (id) ON DELETE CASCADE,
        FOREIGN KEY (ticket_id) REFERENCES tickets (id) ON DELETE CASCADE,
        PRIMARY KEY (project_id, ticket_id));"

           "project_invoices"
           "CREATE TABLE IF NOT EXISTS project_invoices (
        id INTEGER PRIMARY KEY,
        project_id INTEGER NOT NULL,
        client TEXT NOT NULL,
        start_date TEXT NOT NULL,
        end_date TEXT NOT NULL,
        amount INTEGER NOT NULL,
        content TEXT NOT NULL,
        created_on DATETIME DEFAULT CURRENT_TIMESTAMP,
        modified_on DATETIME DEFAULT CURRENT_TIMESTAMP,
        FOREIGN KEY (project_id) REFERENCES projects (id) ON DELETE CASCADE);"}

        7 {"invoices_enhanced"
           "ALTER TABLE invoices ADD COLUMN number TEXT;
       ALTER TABLE invoices ADD COLUMN tax_rate REAL DEFAULT 0.0;
       ALTER TABLE invoices ADD COLUMN payment_terms INTEGER DEFAULT 30;
       ALTER TABLE invoices ADD COLUMN template TEXT DEFAULT 'default';"

           "invoice_templates"
           "CREATE TABLE IF NOT EXISTS invoice_templates (
        id INTEGER PRIMARY KEY,
        name TEXT UNIQUE NOT NULL,
        content TEXT NOT NULL,
        created_on DATETIME DEFAULT CURRENT_TIMESTAMP,
        modified_on DATETIME DEFAULT CURRENT_TIMESTAMP);"

           "invoice_settings"
           "CREATE TABLE IF NOT EXISTS invoice_settings (
        id INTEGER PRIMARY KEY,
        setting TEXT UNIQUE NOT NULL,
        value TEXT NOT NULL,
        created_on DATETIME DEFAULT CURRENT_TIMESTAMP,
        modified_on DATETIME DEFAULT CURRENT_TIMESTAMP);"}

        ;; Add new schema versions here with their changes
        })

;; === Migration Functions ===
(fn db/get-version [db]
  "Get current database version"
  (var version 0)
  (pcall (fn []
           (each [row (db:nrows "SELECT db_ver from version;")]
             (set version (. row "db_ver")))))
  version)

(fn db/set-version [db ver]
  "Set database version"
  (db:exec "DELETE FROM version;")
  (let [stmt (db:prepare "INSERT INTO version (db_ver) VALUES (?);")]
    (stmt:bind_values ver)
    (stmt:step)
    (stmt:finalize)))

(fn db/migrate [db current-ver target-ver]
  "Migrate database from current version to target version"
  (print (.. "DB{Migrating database from version " current-ver " to " target-ver "}"))

  (fn apply-version [ver]
    (let [changes (. schema-versions ver)]
      (db:exec "BEGIN TRANSACTION;")
      (pcall (fn []
               (each [_ stmt (pairs changes)]
                 (db:exec stmt))
               (db/set-version db ver)
               (db:exec "COMMIT;")))))

  (if (< current-ver target-ver)
      (for [ver (+ current-ver 1) target-ver]
        (when (. schema-versions ver)
          (apply-version ver)))))

(fn db/ensure-schema []
  "Ensure database schema is at the latest version"
  (let [db (sql.open conf.tkt_loc)
        latest-ver (accumulate [max-ver 0
                                ver _ (pairs schema-versions)]
                     (if (> ver max-ver) ver max-ver))]
    ;; Create version table if it doesn't exist
    (db:exec (. (. schema-versions 1) "version"))

    ;; Get current version and migrate if needed
    (let [current-ver (db/get-version db)]
      (when (< current-ver latest-ver)
        (db/migrate db current-ver latest-ver)))

    (db:close)))

;; === Replace existing db/defnschema function ===
(fn db/defnschema []
  "Create initial database schema"
  (let [db (sql.open conf.tkt_loc)]
    ;; Apply first version schema
    (each [_ stmt (pairs (. schema-versions 1))]
      (db:exec stmt))
    ;; Set initial version
    (db/set-version db 1)
    (db:close))
  (db/ensure-schema))

;;===== util =====
(fn util/split [val check]
  "Split string on delim and return as table"
  (var check check)
  (if (= val nil)
      []
      (do
        (if (= check nil)
            (set check "%s"))
        (local t {})
        (each [str (string.gmatch val (.. "([^" check "]+)"))]
          (table.insert t str))
        t)))

(fn util/exists? [f]
  "Check if file exists"
  (local f? (io.open f "r"))
  (if f?
      (f?:close)
      false))

(fn util/dir_exists? [dir]
  (let
      [(stat err code) (os.rename dir dir)]
    (when (= stat nil)
      (when (= err 13)
        (lua "return true"))
      (lua "return false"))
    true))

(fn util/mitigate []
  "Ensure dotfile consistency between versions"
  (local valid_vars ["editor" "mail" "rate" "user" "tkt_loc" "invoice_template" "company"])
  (local valid_holds [(os.getenv "EDITOR") "mutt -s" "0.00" (os.getenv "USER") (.. dot.dir "tkts.db") (.. dot.dir "invoice.tf") "Change Me"])
  (var reset nil)
  ;;If there is a value for each setting, leave it be
  (each [k v (pairs valid_vars)]
    (do
      (if (= (. conf v) nil)
          (do
            ;;set holds to match config values
            (tset conf v (. valid_holds k))
            (set reset true)))))
  ;;If a value is missing re-serialize the file
  (if (= reset true)
      (with-open [f (io.open dot.file "w")]
        (f:write (inspect conf)))))

(fn util/conf []
  "Load dotfile configuration"
  ;;if .config/tkts exists
  (if (util/dir_exists? dot.dir)
      (do
        ;;if .config/tkts/tkts.conf exists
        (if (util/exists? dot.file)
            (do
              ;;load config & verify all settings exist
              (with-open [f (io.open dot.file "r")]
                (set conf (lume.deserialize (f:read "*a"))))
              (util/mitigate))
            ;;else, create a new one with defaults
            (with-open [f (io.open dot.file "w")]
              (f:write (inspect conf))))
        ;;if tkts.db exists, check if migration is needed or not, if it doesn't exist create it
        (if (util/exists? conf.tkt_loc)
            (db/ensure-schema) ;;this is just a wrapper around db/migrate
            (db/defnschema))
        (if (util/exists? conf.invoice_template)
            nil
            (with-open [f (io.open conf.invoice_template "w")]
              (f:write (. invoice_templates "default")))))
      (do
        ;;if ~/.config/tkts doesn't exist create it, serialize a new tkts.conf & create a db, and populate groff templates
        (os.execute (.. "mkdir " dot.dir))
        (with-open [f (io.open dot.file "w")]
          (f:write (inspect conf)))
        (with-open [f (io.open conf.invoice_template "w")]
          (f:write (. invoice_templates "default")))
        (db/defnschema))))


;; this should probably be extended so that I can replace the io.sterr:write calls more generally
;;(util/err "tkt" args "A tkt number is required to do X with Y.")
(fn util/err [arg args msg]
  (if (= (. args arg) nil)
      (do
        (io.stderr:write (.. msg "\n"))
        (os.exit))))

;;(util/template {1 "Templates are just strings with {variables}" :variables "Something substituted"})
(fn util/template [str vars]
  "When passed a template and a table of variables, populate it. Otherwise when passed a string and a var, act like ruby's #{var}"
  (when (not vars)
    (let
        [vars str
         str (. vars 1)]
      (string.gsub str "({([^}]+)})"
                   (fn [whole i]
                     (or (. vars i) whole))))))

(fn yaml/to-template [data template-type]
  "Convert data to YAML template based on type"
  (match template-type
    :ticket
    (let [yaml {:metadata {:title data.title
                           :status data.status
                           :owner data.owner
                           :client data.client
                           :project data.project
                           :time data.time}
                :description data.desc}]
      (lyaml.dump [yaml]))

    :record
    (let [yaml {:metadata {:user data.user
                           :time data.time
                           :tag (or data.tag "")}
                :content data.log}]
      (lyaml.dump [yaml]))

    :company
    (let [yaml {:metadata {:name data.name
                           :abbreviation data.abbrv}
                :address data.address}]
      (lyaml.dump [yaml]))

    :client
    (let [yaml {:metadata {:name data.name
                           :company data.company
                           :role data.role
                           :phone data.phone
                           :email data.email}}]
      (lyaml.dump [yaml]))

    :project
    (let [yaml {:metadata {:name data.name
                           :status data.status
                           :client data.client
                           :company data.company}
                :dates {:start_date (or data.start_date "")
                        :end_date (or data.end_date "")}
                :description (or data.description "")}]
      (lyaml.dump [yaml]))))

(fn yaml/from-template [yaml-str template-type]
  "Convert YAML back to table"
  (let [(ok parsed) (pcall lyaml.load yaml-str)]
    (if (not ok)
        (do
          (print "YAML parsing error:" parsed)
          ;; Return empty structure to prevent nil indexing
          (match template-type
            :ticket {:title "" :status "" :owner "" :client "" :project "" :time 0 :desc ""}
            :record {:user "" :time 0 :tag "" :log ""}
            :company {:name "" :abbrv "" :address ""}
            :client {:name "" :company "" :role "" :phone "" :email ""}))

        ;; If parsing succeeded, use the parsed data directly
        (match template-type
          :ticket
          {:title (or (. parsed :metadata :title) "")
           :status (or (. parsed :metadata :status) "")
           :owner (or (. parsed :metadata :owner) "")
           :client (or (. parsed :metadata :client) "")
           :project (or (. parsed :metadata :project) "")
           :time (or (. parsed :metadata :time) 0)
           :desc (or (. parsed :description) "")}

          :record
          {:user (. parsed :metadata :user)
           :time (. parsed :metadata :time)
           :tag (. parsed :metadata :tag)
           :log (. parsed :content)}

          :company
          {:name (. parsed :metadata :name)
           :abbrv (. parsed :metadata :abbreviation)
           :address (. parsed :address)}

          :client
          {:name (. parsed :metadata :name)
           :company (. parsed :metadata :company)
           :role (. parsed :metadata :role)
           :phone (. parsed :metadata :phone)
           :email (. parsed :metadata :email)}

          :project
          {:name (. parsed :metadata :name)
           :status (. parsed :metadata :status)
           :client (. parsed :metadata :client)
           :company (. parsed :metadata :company)
           :start_date (. parsed :dates :start_date)
           :end_date (. parsed :dates :end_date)
           :description (. parsed :description)}))))

;; ===== Command Parser =====
(local commands {})
(local aliases {})

(fn cmd/register-alias [alias-name target-name]
  "Register a command alias"
  (tset aliases alias-name target-name))

(fn cmd/resolve-alias [name]
  "Resolve command alias to actual command name"
  (or (. aliases name) name))

(fn cmd/register [name opts handler]
  "Register a command with its options and handler"
  ;; Split the command name on colons for namespaced commands
  (let [segments (util/split name ":")]
    (var current commands)
    (each [i segment (ipairs segments)]
      (when (= i (length segments))
        (tset current segment {:opts opts :handler handler})
        (lua "return"))
      (when (= (. current segment) nil)
        (tset current segment {}))
      (set current (. current segment)))))

(fn cmd/get [args]
  "Get command and remaining args from input"
  (if (= (length args) 0)
      (values (. commands :default) [:default] [])
      (let [first-arg (. args 1)]
        ;; Special case for help command
        (if (or (= first-arg "help") (= first-arg "--help") (= first-arg "-h"))
            (values (. commands :help) [:help] (table.move args 2 (or (length args) 0) 1 {}))
            ;; Normal command resolution
            (let [cmd-name (cmd/resolve-alias first-arg)
                  cmd-segments (if cmd-name
                                   (util/split cmd-name ":")
                                   ["default"])
                  remaining (if (> (length args) 1)
                                (table.move args 2 (length args) 1 {})
                                {})]
			  (var current commands)

              ;; Navigate through command segments
              (each [_ segment (ipairs cmd-segments)]
                (when current
                  (set current (. current segment))))

              (values current cmd-segments remaining))))))

(fn parse/flag-value [flags i]
  "Get value for a flag, handling both --flag value and --flag=value formats"
  (let [flag (. flags i)
        next-flag (. flags (+ i 1))]
    (if (string.find flag "=")
        (let [(name val) (string.match flag "%-%-(.-)=(.+)")]
          (values name val (+ i 1)))
        (if (and next-flag (not (string.match next-flag "^%-")))
            (values (string.match flag "%-%-(.+)") next-flag (+ i 2))
            (values (string.match flag "%-%-(.+)") true (+ i 1))))))

(fn parse/short-flag [flag]
  "Parse short flag format (-t value or -t=value)"
  (if (string.find flag "=")
      (let [(name val) (string.match flag "%-(.-)=(.+)")]
        (values name val))
      (values (string.match flag "%-(.+)") true)))

(fn cmd/parse-opts [opts flags]
  "Parse command line flags into options table"
  (let [parsed {}
        flag-map {}]
    ;; Build flag map for both long and short forms
    (each [_ opt (ipairs opts)]
      (tset flag-map opt.name opt)
      (when opt.short
        (tset flag-map opt.short opt)))

    (var i 1)
    (while (<= i (length flags))
      (let [flag (. flags i)]
        (if (string.match flag "^%-%-")
            ;; Long flag
            (let [(name val next-i) (parse/flag-value flags i)]
              (let [opt (. flag-map name)]
                (when opt
                  (tset parsed opt.name val))
                (set i next-i)))
            ;; Short flag
            (string.match flag "^%-")
            (let [(name val) (parse/short-flag flag)]
              (let [opt (. flag-map name)]
                (when opt
                  (if (= val true)
                      (do
                        (let [next-flag (. flags (+ i 1))]
                          (if (and next-flag (not (string.match next-flag "^%-")))
                              (do
                                (tset parsed opt.name next-flag)
                                (set i (+ i 2)))
                              (do
                                (tset parsed opt.name true)
                                (set i (+ i 1))))))
                      (do
                        (tset parsed opt.name val)
                        (set i (+ i 1))))))))))
    parsed))

;;===== sql =====
;;(sql/modify (sql.open conf.tkt_loc) 1 "tickets" ["time" "status"] {"time" 123 "status" "Closed"})
(fn sql/modify [db id table coltbl valtbl]
  "Update existing columns with specified values"
  (if (or (~= (type coltbl) "table") (~= (type valtbl) "table"))
      (do
        (io.stderr:write "Invalid table passed to sql/modify\n")
        (os.exit)))
  (var update "")
  ;;collate coltbl values with valtbl values using coltbl as keys
  (each [k v (ipairs coltbl)]
    (set update (.. update (.. "'" v "' = '" (-> (. valtbl v)
                                                 (string.gsub "'" "''")) "',"))))
  ;;:sub 1 -2 here removes the last trailing ,
  (let
      [stmt (db:prepare (.. "UPDATE " table " SET " (update:sub 1 -2) " WHERE id = " id ";"))]
    (stmt:step)
    (stmt:finalize)))

;;(sql/select (sql.open conf.tkt_loc) "client" ["email" "role"] "LIKE" ["name"] ["will"]) <- select where name like will
;;(sql/select (sql.open conf.tkt_loc) "client" ["email" "role"] "LIKE" ["name"] [""]) <- select anything
;;(sql/select (sql.open conf.tkt_loc) "tickets" ["client" "title"] "=" ["id"] [10]) <- select where X = Y
;;(sql/select (sql.open conf.tkt_loc) "company" ["name"] "!=" ["abbrv"] ["LC"]) <- select where X not like Y
(fn sql/select [db sqltbl coltbl fltrtyp fltrtbl valtbl]
  (if (or (~= (type coltbl) "table") (~= (type fltrtbl) "table") (~= (type valtbl) "table"))
      (do
        (io.stderr:write "Invalid table passed to sql/select.\n")
        (os.exit))
      (and (~= fltrtyp "LIKE") (~= fltrtyp "=") (~= fltrtyp "!="))
      (do
        (io.stderr:write (.. fltrtyp " is an invalid filter type. Try LIKE, =, or !=.\n"))
        (os.exit)))
  (var selection "")
  (var filter "")
  (each [k v (ipairs coltbl)]
    (set selection (.. selection (.. "\"" v "\","))))
  (if (> (length fltrtbl) 1)
      (each [k v (ipairs fltrtbl)]
        (set filter (.. filter (.. "\"" (. valtbl k) "\","))))
      (set filter (.. filter (.. (. fltrtbl 1) " " fltrtyp " '" (if (= fltrtyp "LIKE") "%" "") (. valtbl 1) (if (= fltrtyp "LIKE") "%" "") "'"))))
  ;;:sub 1 -2 here removes the last trailing ,
  ;;(print (.. "SELECT " (selection:sub 1 -2) " FROM " sqltbl " WHERE " filter ";"))
  (let
      [stmt (db:prepare (.. "SELECT " (selection:sub 1 -2) " FROM " sqltbl " WHERE " filter ";"))
       retval []]
    (each [row (stmt:nrows)]
      (table.insert retval row))
    (stmt:finalize)
    retval))

;;===== company =====
;;(comp/add {"name" "Lambdacreate" "abbreviation" "LC" "address" "123 Nihil et Nowhere"})
(fn comp/add [args]
  "Add a company"
  (util/err "name" args "You must provide a name to add a company")
  (util/err "abbreviation" args "You must provide an abbreviation to add a company")
  (let
      [db (sql.open conf.tkt_loc)
       stmt (db:prepare "INSERT into company (name,abbrv,address) VALUES (:name, :abbrv, :address)")]
    (stmt:bind_names {:name args.name :abbrv args.abbreviation :address args.address})
    (stmt:step)
    (stmt:finalize))
  ;;FIXME: Change format to C# {Status: Created}
  (print (.. "{:Company " args.name "}")))

;;(comp/edit {"name" "LambdaCreate"})
(fn comp/edit [args]
  "Edit company details using YAML interface"
  (util/err "name" args "A company name is required to edit a company record.")
  (let [db (sql.open conf.tkt_loc)]
    (var found false)
    (each [row (db:nrows (.. "SELECT * from company WHERE name = '" args.name "';"))]
      (set found true)
      (let [company {"name" (. row "name")
                     "abbrv" (. row "abbrv")
                     "address" (if (= (. row "address") nil) "" (. row "address"))}
            id (. row "id")
            tmpf (os.tmpname)]

        ;; Write YAML template to temp file
        (with-open [f (io.open tmpf "w")]
          (f:write (yaml/to-template company :company)))

        ;; Open in editor
        (os.execute (.. conf.editor " " tmpf))

        ;; Read and parse updated YAML
        (with-open [f (io.open tmpf "r")]
          (let [diff (yaml/from-template (f:read "*a") :company)]
            (sql/modify db id "company" ["name" "abbrv" "address"] diff)))

        (print (.. "C" id " {Status: Modified}"))))

    (when (not found)
      (io.stderr:write (.. "No company found matching '" args.name "'.\n"))
      (os.exit))

    (db:close)))

;;(comp/list)
(fn comp/list []
  "Briefly list all companies"
  (let
      [db (sql.open conf.tkt_loc)]
    (each [row (db:nrows "SELECT * FROM company;")]
      (print (.. "{:Id " (. row "id") " :Name " (. row "name") "}")))
    (db:close)))

;;(comp/info "LambdaCreate")
(fn comp/info [name]
  "Return company as lua table"
  (var export {})
  (let [db (sql.open conf.tkt_loc)]
    (each [row (db:nrows (.. "SELECT * from company where name = '" name "' OR abbrv COLLATE NOCASE = '" name "';"))]
      (let [company {"id" (. row "id")
                     "name" (. row "name")
                     "abbrv" (. row "abbrv")
                     "address" (. row "address")
                     "created_on" (. row "created_on")}]
        (set export company))))
  export)

;;===== client =====
;;TODO: Redo this function so it relies on editing a lua table via a tmpf
;;(clie/add "Will" "LambdaCreate" "{role='Dev',phone='123-456-7890',email='wpsinatra@gmail.com'}")
(fn clie/add [args detail]
  "Add a client"
  (util/err "name" args "You must provide a name to add a client.")
  (util/err "company" args "You must provide a company to associate the client with.")
  ;;figure out a way to catch this error, so that you can io.stderr instead of getting an <eof> if you get a malformed table
  (let
      [details (if (~= detail nil) (lume.deserialize detail) "")]
    (if (and (~= detail nil) (~= (type details) "table"))
        (do
          (io.stderr:write "Details must be provided as a table.\nIE: \"{role='X', phone='123-456-7890', email='nihil@nothing'}\n")
          (os.exit))
        (do
          (let
              [db (sql.open conf.tkt_loc)
               stmt (db:prepare "INSERT into client (name,company,role,phone,email) VALUES (:name, :company, :role, :phone, :email)")]
            (stmt:bind_names {:name args.name :company args.company :role details.role :phone details.phone :email details.email})
            (stmt:step)
            (stmt:finalize))
          ;;FIXME: change format to CL# {:Status Created}
          (print (.. "{:Client " args.name "}"))))))

;;(clie/edit {"name" "Will Sinatra" "company" "LambdaCreate"})
(fn clie/edit [args]
  "Edit client details using YAML interface"
  (util/err "name" args "A name is required to edit a client record.")
  (util/err "company" args "The company associated with the user must be provided.")
  (let [db (sql.open conf.tkt_loc)]
    (var found false)
    (each [row (db:nrows (.. "SELECT * FROM client WHERE name = '" args.name "' AND company = '" args.company "';"))]
      (set found true)
      (let [client {"name" (. row "name")
                    "company" (. row "company")
                    "role" (if (= (. row "role") nil) "" (. row "role"))
                    "phone" (if (= (. row "phone") nil) "" (. row "phone"))
                    "email" (if (= (. row "email") nil) "" (. row "email"))}
            id (. row "id")
            tmpf (os.tmpname)]

        ;; Write YAML template to temp file
        (with-open [f (io.open tmpf "w")]
          (f:write (yaml/to-template client :client)))

        ;; Open in editor
        (os.execute (.. conf.editor " " tmpf))

        ;; Read and parse updated YAML
        (with-open [f (io.open tmpf "r")]
          (let [diff (yaml/from-template (f:read "*a") :client)]
            (sql/modify db id "client"
                        ["name" "company" "role" "phone" "email"]
                        diff)))

        (print (.. "CL" id " {Status: Modified}"))))

    (when (not found)
      (io.stderr:write (.. "No client found matching '" args.name "' in company '" args.company "'.\n"))
      (os.exit))

    (db:close)))

;;(clie/list)
(fn clie/list []
  "Briefly list all clients"
  (let
      [db (sql.open conf.tkt_loc)]
    (each [row (db:nrows "SELECT * FROM client;")]
      (print (.. "{:Id " (. row "id") " :Name " (. row "name") " :Comp " (. row "company") "}")))
    (db:close)))

;;(clie/info "Will Sinatra" "LambdaCreate")
(fn clie/info [name company]
  "Return client as lua table"
  (var export {})
  (let
      [db (sql.open conf.tkt_loc)]
    (each [row (db:nrows (.. "SELECT * from client where name = '" name "' AND company LIKE '" company "';"))]
      (let
          [client {"id" (. row "id")
                   "name" (. row "name")
                   "company" (. row "company")
                   "role" (if (= (. row "role") nil) "" (. row "role"))
                   "phone" (if (= (. row "phone") nil) "" (. row "phone"))
                   "email" (if (= (. row "email") nil) "" (. row "email"))
                   "created_on" (. row "created_on")}]
        (set export client))))
  export)

;;===== rec =====
(fn rec/info [tnum amount]
  "Return a table of records"
  (let [db (sql.open conf.tkt_loc)
        record {}
        statement (if (= amount nil)
                      (.. "SELECT * from records where tkt_num = " tnum " ORDER BY id ASC;")
                      (.. "SELECT * from records where tkt_num = " tnum " ORDER BY id DESC limit " amount ";"))]
    (each [row (db:nrows statement)]
      (let [rec {"id" (. row "id")
                 "log" (. row "log")
                 "user" (. row "user")
                 "time" (. row "time")
                 "tag" (if (= (. row "tag") nil) "" (. row "tag"))
                 "created_on" (. row "created_on")}]
        (table.insert record rec)))
    record))

(fn rec/get-tagged-records [tnum tag]
  "Return records for a specific ticket and tag"
  (let [db (sql.open conf.tkt_loc)
        records []
        stmt (db:prepare "SELECT id, time, tag, created_on FROM records
                         WHERE tkt_num = ? AND tag = ?
                         ORDER BY id ASC")]
    (stmt:bind_values tnum tag)
    (each [row (stmt:nrows)]
      (table.insert records row))
    (stmt:finalize)
    (db:close)
    records))

(fn rec/sum-tagged-time [tnum tag]
  "Get total time for a specific tag in a ticket"
  (let [db (sql.open conf.tkt_loc)
        stmt (db:prepare "SELECT SUM(time) as total FROM records
                         WHERE tkt_num = ? AND tag = ?")]
    (stmt:bind_values tnum tag)
    (var total 0)
    (each [row (stmt:nrows)]
      (set total (or (. row "total") 0)))
    (stmt:finalize)
    (db:close)
    total))

(fn rec/summary-by-tag [tnum]
  "Return a summary of time spent per tag for a ticket"
  (let [db (sql.open conf.tkt_loc)
        summary {}]
    (each [row (db:nrows (.. "SELECT tag, SUM(time) as total_time, COUNT(*) as count
                             FROM records
                             WHERE tkt_num = " tnum "
                             AND tag IS NOT NULL
                             GROUP BY tag;"))]
      (tset summary (. row "tag")
            {:time (. row "total_time")
             :count (. row "count")}))
    (db:close)
    summary))

;;===== tkts =====
;;(tkts/count (sql.open conf.tkt_loc))
;;(tkts/count db "Total") ;;or Closed or Open
(fn tkts/count [db ?filter]
  "Return a count of tickets"
  (let
      [sql "SELECT count(id) FROM tickets "
       open "WHERE status = 'Open';"
       closed "WHERE status = 'Closed';"
       count {"open" 0 "closed" 0 "total" 0}
       query (fn [sql targ] (each [row (db:nrows sql)] (tset count targ (. row "count(id)"))))]
    (if (or (= ?filter nil) (= ?filter "Total"))
        (do
          (query (.. sql open) "open")
          (query (.. sql closed) "closed")
          (tset count "total" (+ count.open count.closed)))
        (= ?filter "Closed")
        (query (.. sql closed) "closed")
        (= ?filter "Open")
        (query (.. sql open) "open"))
    count))

;;(tkts/create {"issue" "Some problem" "description" "Some detils"})
(fn tkts/create [args]
  "Create a new open ticket"
  (util/err "issue" args "An issue title is required to create a tkt.")
  (let
      [db (sql.open conf.tkt_loc)
       status "Open"
       time 0
       stmt (db:prepare "INSERT into tickets (status,title,desc,owner,time) VALUES (:status, :title, :desc, :owner, :time)")]
    (stmt:bind_names {:status status :title args.issue :desc args.description :owner conf.user :time time})
    (stmt:step)
    (stmt:finalize)
    (let
        [tid (db:last_insert_rowid)]
      (db:close)
      (print (.. "T" tid " {:Status Open}")))))

;;(tkts/close {"tkt" 1})
(fn tkts/close [args]
  "Move a ticket from Open to Closed"
  (util/err "tkt" args "A tkt number is required to close a tkt.")
  (let
      [db (sql.open conf.tkt_loc)]
    (sql/modify db args.tkt "tickets" ["status"] {"status" "Closed"})
    (db:close))
  (print (.. "T" args.tkt " {:Status Closed}")))

;;(tkts/reopen {"tkt" 1})
(fn tkts/reopen [args]
  "Move a ticket from Closed to Open"
  (util/err "tkt" args "A tkt number is required to reopen a tkt.")
  (let
      [db (sql.open conf.tkt_loc)]
    (sql/modify db args.tkt "tickets" ["status"] {"status" "Open"})
    (db:close))
  (print (.. "T" args.tkt " {:Status Open}")))

;;(tkts/info {"tkt" 1})
;;Add a way to verify an integer is passed as tnum
(fn tkts/info [tnum]
  "Return tkt as lua table"
  (var export {})
  (let
      [db (sql.open conf.tkt_loc)]
    (each [row (db:nrows (.. "SELECT * from tickets where id = " tnum ";"))]
      (let
          [tkt {"status" (. row "status")
                "title" (. row "title")
                "desc" (if (= (. row "desc") nil) "" (. row "desc"))
                "client" (if (= (. row "client") nil) "" (. row "client"))
                "project" (if (= (. row "project") nil) "" (. row "project"))
                "owner" (. row "owner")
                "time" (. row "time")
                "rate" 0
                "log" 0}]
        (tset tkt "rate" (math.ceil (+ (* (/ tkt.time 60) (. conf "rate")) 0.5)))
        (each [row (db:nrows (.. "SELECT count(id) from records where tkt_num = " tnum ";"))]
          (tset tkt "log" (. row "count(id)")))
        (set export tkt))))
  export)

(fn tkts/comment [args]
  "Add a comment to the ticket log via editor, or if passed a string, directly into the ticket log"
  (util/err "tkt" args "A tkt number is required to add logging to a tkt.")
  (let [db (sql.open conf.tkt_loc)
        stmt (db:prepare "INSERT into records (tkt_num, log, user, time, tag) VALUES (:tkt_num, :log, :user, :time, :tag)")]
    (if (= args.log nil)
        (do
          (let [tmpf (os.tmpname)]
            (os.execute (.. conf.editor " " tmpf))
            (with-open [f (io.open tmpf "r")]
              (stmt:bind_names {:tkt_num args.tkt
                                :log (f:read "*a")
                                :user conf.user
                                :time 0
                                :tag args.tag})
              (stmt:step)
              (stmt:finalize)))
          (let [rid (db:last_insert_rowid)]
            rid))
        (do
          (let [info (tkts/info args.tkt)]
            (stmt:bind_names {:tkt_num args.tkt
                              :log args.log
                              :user conf.user
                              :time 0
                              :tag args.tag})
            (stmt:step)
            (stmt:finalize)
            (print (.. "T" args.tkt " {:log "  (+ info.log 1) "}")))))))

;;Add an add record as well, so that per log timing can be added manually.
;;(tkts/time {"tkt" 1})
(fn tkts/time [args]
  "Grab a time stamp, open the ticket's log, append to it, and then append the calculated time spent appending to the ticket log, to the ticket's total time in minutes."
  (util/err "tkt" args "A tkt number is required to record time on a tkt.")
  (let [db (sql.open conf.tkt_loc)
        time {"start" (os.time)
              "end" 0
              "diff" 0
              "cur" 0}]
    (each [row (db:nrows (.. "SELECT time from tickets WHERE id = " args.tkt ";"))]
      (tset time "cur" (. row "time")))
    (let [rnum (tkts/comment (collect [k v (pairs args)] k v)) ; Preserve any tags passed in args
          info (tkts/info args.tkt)
          tnum (if (= info.log 0) 1 info.log)]
      (tset time "end" (os.time))
      (tset time "diff" (math.ceil (+ (/ (os.difftime time.end time.start) 60) 0.5)))
      (sql/modify db args.tkt "tickets" ["time"] {"time" (+ time.cur time.diff)})
      (sql/modify db rnum "records" ["time"] {"time" time.diff})
      (db:close)
      (print (.. "T" args.tkt " {prev: " time.cur ", cur: " (+ time.cur time.diff) ", log: " tnum
                 (if args.tag (.. ", tag: " args.tag) "") "}")))))

;;(tkts/list "Open")
(fn tkts/list [status]
  "List tickets by status"
  (let
      [db (sql.open conf.tkt_loc)]
    (print (.. status " Tickets:"))
    (each [row (db:nrows (.. "SELECT * from tickets WHERE status = '" status "';"))]
      (print (.. (.. "T" row.id ") ") row.title)))
    (db:close)))

;;(tkts/edit/tkt 1)
(fn tkts/edit/tkt [tnum]
  "Edit ticket using YAML interface"
  (let [db (sql.open conf.tkt_loc)]
    (each [row (db:nrows (.. "SELECT * from tickets WHERE id = '" tnum "';"))]
      (let [tkt {"status" (. row "status")
                 "title" (. row "title")
                 "desc" (if (= (. row "desc") nil) "" (. row "desc"))
                 "client" (if (= (. row "client") nil) "" (. row "client"))
                 "project" (if (= (. row "project") nil) "" (. row "project"))
                 "owner" (. row "owner")
                 "time" (. row "time")}
            tmpf (os.tmpname)]
        ;; Convert to YAML and write to temp file
        (with-open [f (io.open tmpf "w")]
          (f:write (yaml/to-template tkt :ticket)))

        ;; Open in editor
        (os.execute (.. conf.editor " " tmpf))

        ;; Read back and parse YAML
        (with-open [f (io.open tmpf "r")]
          (let [diff (yaml/from-template (f:read "*a") :ticket)]
            (sql/modify db tnum "tickets"
                        ["client" "desc" "owner" "project" "status" "time" "title"]
                        diff)))))
    (db:close)
    (print (.. "T" tnum " {:Status Modified}"))))

;;(log/edit 1 123)
(fn tkts/edit/log [tnum log]
  "Edit record using YAML interface"
  (let [db (sql.open conf.tkt_loc)
        lnum (tonumber log)
        recs (sql/select db "records" ["id" "log" "user" "time" "tag"] "=" ["tkt_num"] [tnum])
        record {"log" (. recs lnum "log")
                "user" (. recs lnum "user")
                "time" (. recs lnum "time")
                "tag" (if (= (. recs lnum "tag") nil) "" (. recs lnum "tag"))}
        tmpf (os.tmpname)]

    ;; Write YAML template to temp file
    (with-open [f (io.open tmpf "w")]
      (f:write (yaml/to-template record :record)))

    ;; Open in editor
    (os.execute (.. conf.editor " " tmpf))

    ;; Read and parse updated YAML
    (with-open [f (io.open tmpf "r")]
      (let [diff (yaml/from-template (f:read "*a") :record)]
        (sql/modify db (. recs lnum "id") "records"
                    ["log" "user" "time" "tag"]
                    diff)))

    ;; Update total time on ticket
    (each [row (db:nrows (.. "SELECT sum(time) from records WHERE tkt_num = " tnum ";"))]
      (sql/modify db tnum "tickets" ["time"] {"time" (. row "sum(time)")}))

    (db:close))
  (print (.. "T" tnum " {:log " log " :Status Modified}")))

(fn tkts/edit [args]
  (util/err "tkt" args "A tkt number is required to edit a tkt log.")
  (if (= args.log nil)
      (tkts/edit/tkt args.tkt)
      (tkts/edit/log args.tkt args.log)))

(fn tkts/prepare-for-invoice [args]
  "Prepare a ticket for invoicing by setting required fields"
  (util/err "tkt" args "A tkt number is required.")
  (util/err "client" args "A client name is required.")

  (let [tkt (tkts/info args.tkt)]
    ;; Validate ticket exists
    (when (= tkt {})
      (io.stderr:write (.. "Ticket T" args.tkt " not found.\n"))
      (os.exit))

    ;; Validate client exists
    (let [client (clie/info args.client "%")]
      (when (not client)
        (io.stderr:write (.. "Client '" args.client "' not found. Create client first with:\n"
                             "tkts client:create -n \"" args.client "\" -c COMPANY\n"))
        (os.exit))

      ;; Update ticket with client
      (let [db (sql.open conf.tkt_loc)]
        (sql/modify db args.tkt "tickets" ["client"] {"client" args.client})
        (db:close))

      (print (.. "Updated T" args.tkt " with client " args.client)))))

;;===== mail =====
;;(tkts/mail {"tkt" 1})
(fn tkts/mail [args]
  (let
      [db (sql.open conf.tkt_loc)
       tkt (sql/select db "tickets" ["title" "client"] "=" ["id"] [args.tkt])
       mail (sql/select db "client" ["email"] "=" ["name"] [(. tkt 1 "client")])]
    (os.execute (.. conf.mail " '" (. tkt 1 "title") "' " (. mail 1 "email")))))

;;;;===== invoice =====
(fn inv/format-currency [amount]
  "Format amount as currency"
  (string.format "%.2f" (or amount 0)))

(fn inv/format-hours [minutes]
  "Convert minutes to hours with 2 decimal precision"
  (string.format "%.2f" (/ minutes 60.0)))

(fn inv/format-ticket-table [tkt-num title line-items]
  "Format a complete ticket table with header and items"
  (let [template-vars {:tkt_num tkt-num
                       :tkt_title title
                       :line_items (table.concat line-items "\n")}
        template-table (collect [k v (pairs template-vars)] k v)]
    (tset template-table 1 ticket-table-template)
    (util/template template-table)))

(fn inv/format-line-item [item]
  "Format line item for invoice template"
  (string.format "%s@%.2f@@"
                 item.desc
                 item.hours))

(fn inv/format-line-items [groups rate]
  "Format all line items with company subtotals"
  (var current-company nil)
  (var company-total-minutes 0)
  (var lines [])

  (let [sorted-keys (lume.keys groups)
        _ (table.sort sorted-keys)]

    (each [_ key (ipairs sorted-keys)]
      (let [group (. groups key)
            company group.company
            hours (inv/format-hours group.minutes)]

        ;; Add line item
        (table.insert lines (inv/format-line-item {:desc key :hours hours}))

        ;; If company changes or this is the last item, add total
        (if (and current-company (not= company current-company))
            (do
              ;; Add previous company total
              (table.insert lines (string.format "@%.2f@$%.2f@$%.2f"
                                                 (inv/format-hours company-total-minutes)
                                                 rate
                                                 (* rate (/ company-total-minutes 60))))
              (set company-total-minutes group.minutes))
            ;; Same company, accumulate minutes
            (set company-total-minutes (+ (or company-total-minutes 0) group.minutes)))

        (set current-company company)))

    ;; Add final company total if there are any records
    (when current-company
      (table.insert lines (string.format "@%.2f@$%.2f@$%.2f"
                                         (inv/format-hours company-total-minutes)
                                         rate
                                         (* rate (/ company-total-minutes 60)))))

    (table.concat lines "\n")))

(fn inv/get-abbreviations [db]
  "Get mapping of company names to abbreviations"
  (let [abbrvs {}]
    (each [row (db:nrows "SELECT name, abbrv FROM company;")]
      (tset abbrvs (. row "name") (. row "abbrv")))
    abbrvs))

(fn inv/group-by-abbrv [records companies]
  "Group records by company abbreviation and tag"
  (let [groups {}]
    (each [_ record (ipairs records)]
      (let [company-abbrv (or (. companies (. record "company")) "UNKNOWN")
            tag (if (or (= record.tag nil) (= record.tag ""))
                    "General"
                    record.tag)
            key (.. "[" company-abbrv "] - " tag)]
        (if (. groups key)
            (tset groups key {:minutes (+ (. (. groups key) :minutes) record.time)
                              :company company-abbrv})
            (tset groups key {:minutes record.time
                              :company company-abbrv}))))
    groups))

(fn inv/group-by-tag [records]
  "Group records by tag and calculate total time"
  (let [groups {}]
    (each [_ record (ipairs records)]
      (let [tag (if (or (= record.tag nil) (= record.tag ""))
                    "General"
                    record.tag)
            time (/ record.time 60.0)]
        (if (. groups tag)
            (tset (. groups tag) :time
                  (+ (. (. groups tag) :time) time))
            (tset groups tag {:time time
                              :tag tag}))))
    groups))

(fn inv/get-records [tnum start-date end-date]
  "Get records for a ticket within date range"
  (let [db (sql.open conf.tkt_loc)
        stmt (db:prepare "SELECT * FROM records
                         WHERE tkt_num = ?
                         AND created_on >= ?
                         AND created_on <= ?
                         ORDER BY created_on ASC")
        records []]
    (stmt:bind_values tnum start-date end-date)
    (each [row (stmt:nrows)]
      (table.insert records row))
    (stmt:finalize)
    (db:close)
    records))

(fn inv/validate-requirements [tkt]
  "Validate all requirements are met before generating an invoice"
  (let [requirements []]

    ;; Check user exists as client
    (when (not (clie/info conf.user "%"))
      (table.insert requirements (.. "User '" conf.user "' must be set up as a client first. Run:\n"
                                     "  1. tkts company:create -n \"" conf.company "\" --abv ABBREVIATION --addr \"YOUR ADDRESS\"\n"
                                     "  2. tkts client:create -n \"" conf.user "\" -c ABBREVIATION\n")))

    ;; Check user's company exists
    (when (not (comp/info conf.company))
      (table.insert requirements (.. "Company '" conf.company "' not found. Run:\n"
                                     "  tkts company:create -n \"" conf.company "\" --abv ABBREVIATION --addr \"YOUR ADDRESS\"")))

    ;; Check ticket has a client assigned
    (when (or (= tkt.client nil) (= tkt.client ""))
      (table.insert requirements "Ticket must be assigned to a client (use tkts edit -t NUMBER to set client)"))

    ;; Check client exists
    (when (and tkt.client (not (clie/info tkt.client "%")))
      (table.insert requirements (.. "Client '" tkt.client "' not found in database")))

    ;; If client exists, check their company exists
    (when tkt.client
      (let [client (clie/info tkt.client "%")]
        (when (and client (not (comp/info (. client "company"))))
          (table.insert requirements (.. "Company '" (. client "company") "' not found in database")))))

    ;; Check ticket has time logged
    (when (= tkt.time 0)
      (table.insert requirements "No time logged on ticket (use tkts time -t NUMBER to log time)"))

    requirements))

(fn inv/check-prerequisites [args]
  "Check and validate all prerequisites for invoice generation"
  (util/err "tkt" args "A tkt number is required to generate an invoice.")

  ;; Get ticket info
  (let [tkt (tkts/info args.tkt)]
    (if (= tkt {})
        (do
          (io.stderr:write (.. "Ticket T" args.tkt " not found.\n"))
          (os.exit))
        (let [requirements (inv/validate-requirements tkt)]
          (when (> (length requirements) 0)
            (io.stderr:write "Cannot generate invoice. The following requirements must be met:\n")
            (each [_ req (ipairs requirements)]
              (io.stderr:write (.. "- " req "\n")))
            (os.exit))))))

;; Invoice generation
(fn inv/generate-number [prefix date]
  "Generate invoice number using prefix and date"
  (let [timestamp (os.time (or date (os.date "*t")))
        date-str (os.date "%m%d%y" timestamp)]
    (.. prefix date-str)))

(fn inv/calculate-totals [line-items tax-rate]
  "Calculate subtotal, tax and final amount"
  (let [subtotal (accumulate [total 0
                              _ item (ipairs line-items)]
                   (+ total (* item.hours item.rate)))
        tax-amount (* subtotal (/ tax-rate 100))
        final-amount (+ subtotal tax-amount)]
    {:subtotal subtotal
     :tax_amount tax-amount
     :final_amount final-amount}))

(fn inv/store [db invoice]
  "Store complete invoice record"
  (let [stmt (db:prepare "INSERT INTO invoices
                         (number, tkt_num, client, content, amount,
                          start_date, end_date, tax_rate, payment_terms)
                         VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)")
        _ (stmt:bind_values invoice.number
                            invoice.tkt_num
                            invoice.client
                            invoice.content
                            invoice.final_amount
                            invoice.start_date
                            invoice.end_date
                            invoice.tax_rate
                            invoice.payment_terms)
        _ (stmt:step)
        id (db:last_insert_rowid)]
    (stmt:finalize)
    id))

(fn inv/update [db id invoice]
  "Update existing invoice with new content and amounts"
  (let [stmt (db:prepare "UPDATE invoices
                         SET content = ?,
                             amount = ?,
                             tax_rate = ?,
                             payment_terms = ?,
                             modified_on = CURRENT_TIMESTAMP
                         WHERE id = ?")]
    (stmt:bind_values invoice.content
                      invoice.final_amount
                      invoice.tax_rate
                      invoice.payment_terms
                      id)
    (stmt:step)
    (stmt:finalize)))

(fn inv/get-by-client [client]
  "Get all invoices for a client"
  (let [db (sql.open conf.tkt_loc)
        stmt (db:prepare "SELECT * FROM invoices WHERE client = ? ORDER BY created_on DESC")
        invoices []]
    (stmt:bind_values client)
    (each [row (stmt:nrows)]
      (table.insert invoices row))
    (stmt:finalize)
    (db:close)
    invoices))

(fn inv/get-by-ticket [tnum]
  "Get all invoices for a ticket"
  (let [db (sql.open conf.tkt_loc)
        stmt (db:prepare "SELECT * FROM invoices WHERE tkt_num = ? ORDER BY created_on DESC")
        invoices []]
    (stmt:bind_values tnum)
    (each [row (stmt:nrows)]
      (table.insert invoices row))
    (stmt:finalize)
    (db:close)
    invoices))

(fn tkts/invoice [args]
  "Generate an invoice for a ticket with detailed line items"
  ;; Check prerequisites
  (inv/check-prerequisites args)

  (let [db (sql.open conf.tkt_loc)
        start-date (or args.start (os.date "%Y-%m-%d" 0))
        end-date (or args.end (os.date "%Y-%m-%d"))
        tax-rate (or args.tax_rate invoice-defaults.tax_rate)
        payment-terms (or args.payment_terms invoice-defaults.payment_terms)

        ;; Get ticket info and records
        tkt (tkts/info args.tkt)
        records (inv/get-records args.tkt start-date end-date)
        companies (inv/get-abbreviations db)

        ;; Group records and format line items
        groups (inv/group-by-abbrv records companies)
        line-items (inv/format-line-items groups (tonumber conf.rate))

        ;; Calculate totals
        total-minutes (accumulate [total 0
                                   _ record (ipairs records)]
                        (+ total record.time))
        total-amount (* (/ total-minutes 60) (tonumber conf.rate))
        tax-amount (* total-amount (/ tax-rate 100))
        final-amount (+ total-amount tax-amount)

        ;; Get user and company info
        user (clie/info conf.user conf.company)
        user_company (comp/info (. user "company"))
        client (clie/info tkt.client "%")
        company (comp/info (. client "company"))
        address (util/split (. company "address") ",")

        ;; Generate invoice number and dates
        number (inv/generate-number (.. "T" args.tkt) nil)
        issued-date (os.date "%x")
        due-date (os.date "%x" (+ (os.time) (* payment-terms 24 60 60)))
        period (.. start-date "-" end-date)

        ;; Format template variables
        template-vars (let [vars {:full_name (. user "name")
                                  :address (. user_company "address")
                                  :cell_number (. user "phone")
                                  :company (. client "company")
                                  :company_addr (.. (. address 1) "," (. address 2))
                                  :company_zip (string.gsub (. address 3) "^ " "")
                                  :client_email (. client "email")
                                  :client_cellphone (. client "phone")
                                  :invoice_number number
                                  :issued_date issued-date
                                  :due_date due-date
                                  :period period
                                  :line_items line-items
                                  :tax_rate tax-rate
                                  :total_cost (inv/format-currency total-amount)
                                  :tax_amount (inv/format-currency tax-amount)
                                  :final_amount (inv/format-currency final-amount)}]
                        (collect [k v (pairs vars)] k v))

        ;; Add template to vars and generate content
        _ (tset template-vars 1 (. invoice_templates (or args.template :default)))
        content (util/template template-vars)

        ;; Store invoice
        invoice {:number number
                 :tkt_num args.tkt
                 :client client.company
                 :content content
                 :start_date start-date
                 :end_date end-date
                 :tax_rate tax-rate
                 :payment_terms payment-terms
                 :final_amount final-amount}
        id (if args.update
               (do
                 (inv/update db args.update invoice)
                 args.update)
               (inv/store db invoice))]

    (print content)
    (db:close)))

;;===== projects =====
(fn proj/create [args]
  "Create a new project"
  (util/err "name" args "A project name is required to create a project.")
  (let [db (sql.open conf.tkt_loc)
        stmt (db:prepare "INSERT INTO projects (name, description, client, company)
                         VALUES (:name, :description, :client, :company)")]
    (stmt:bind_names {:name args.name
                      :description args.description
                      :client args.client
                      :company args.company})
    (stmt:step)
    (stmt:finalize)
    (let [pid (db:last_insert_rowid)]
      (db:close)
      (print (.. "P" pid " {:Status Active}")))))

(fn proj/edit [args]
  "Edit project details using YAML interface"
  (util/err "project" args "A project ID is required to edit a project.")
  (let [db (sql.open conf.tkt_loc)]
    (each [row (db:nrows (.. "SELECT * from projects WHERE id = " args.project ";"))]
      (let [project {"name" (. row "name")
                     "description" (if (= (. row "description") nil) "" (. row "description"))
                     "status" (. row "status")
                     "client" (if (= (. row "client") nil) "" (. row "client"))
                     "company" (if (= (. row "company") nil) "" (. row "company"))
                     "start_date" (if (= (. row "start_date") nil) "" (. row "start_date"))
                     "end_date" (if (= (. row "end_date") nil) "" (. row "end_date"))}
            id (. row "id")
            tmpf (os.tmpname)]

        ;; Write YAML template to temp file
        (with-open [f (io.open tmpf "w")]
          (f:write (yaml/to-template project "project")))

        ;; Open in editor
        (os.execute (.. conf.editor " " tmpf))

        ;; Read and parse updated YAML
        (with-open [f (io.open tmpf "r")]
          (let [diff (yaml/from-template (f:read "*a") "project")]
            (sql/modify db id "projects"
                        ["name" "description" "status" "client" "company"
                         "start_date" "end_date"]
                        diff)))))

    (db:close)
    (print (.. "P" args.project " {:Status Modified}"))))

(fn proj/add-ticket [args]
  "Add a ticket to a project and update ticket with project details"
  (util/err "project" args "A project ID is required.")
  (util/err "ticket" args "A ticket ID is required.")
  (let [db (sql.open conf.tkt_loc)
        ;; First get project info
        proj-stmt (db:prepare "SELECT name, client, company FROM projects WHERE id = ?")
        _ (proj-stmt:bind_values args.project)]
    (var project nil)

    ;; Get project details
    (each [row (proj-stmt:nrows)]
      (set project {:name (. row "name")
                    :client (. row "client")
                    :company (. row "company")}))
    (proj-stmt:finalize)

    (if project
        (do
          ;; Add ticket to project
          (let [link-stmt (db:prepare "INSERT INTO project_tickets (project_id, ticket_id)
                                     VALUES (?, ?)")]
            (link-stmt:bind_values args.project args.ticket)
            (link-stmt:step)
            (link-stmt:finalize))

          ;; Update ticket with project info
          (let [update-stmt (db:prepare "UPDATE tickets
                                       SET project = ?, client = ?
                                       WHERE id = ?")]
            (update-stmt:bind_values project.name project.client args.ticket)
            (update-stmt:step)
            (update-stmt:finalize))

          (db:close)
          (print (.. "P" args.project " {:Status Modified :add T" args.ticket "}")))

        (do
          (db:close)
          (io.stderr:write (.. "Project P" args.project " not found.\n"))
          (os.exit)))))

(fn proj/remove-ticket [args]
  "Remove a ticket from a project and clear project-related fields"
  (util/err "project" args "A project ID is required.")
  (util/err "ticket" args "A ticket ID is required.")
  (let [db (sql.open conf.tkt_loc)
        ;; First verify the ticket is actually in this project
        check-stmt (db:prepare "SELECT 1 FROM project_tickets
                               WHERE project_id = ? AND ticket_id = ?")]
    (check-stmt:bind_values args.project args.ticket)
    (var found false)
    (each [_ (check-stmt:nrows)]
      (set found true))
    (check-stmt:finalize)

    (if found
        (do
          ;; Remove project link
          (let [unlink-stmt (db:prepare "DELETE FROM project_tickets
                                        WHERE project_id = ? AND ticket_id = ?")]
            (unlink-stmt:bind_values args.project args.ticket)
            (unlink-stmt:step)
            (unlink-stmt:finalize))

          ;; Clear project and client fields from ticket
          (let [clear-stmt (db:prepare "UPDATE tickets
                                      SET project = '', client = ''
                                      WHERE id = ?")]
            (clear-stmt:bind_values args.ticket)
            (clear-stmt:step)
            (clear-stmt:finalize))

          (db:close)
          (print (.. "P" args.project " {:Status Modified :rm T" args.ticket "}")))

        (do
          (db:close)
          (io.stderr:write (.. "Ticket T" args.ticket " is not in project P" args.project ".\n"))
          (os.exit)))))

(fn proj/get-ticket-hours [tickets start-date end-date]
  "Get hours worked on tickets within date range"
  (let [db (sql.open conf.tkt_loc)
        hours {}]
    (each [_ ticket (ipairs tickets)]
      (let [stmt (db:prepare "SELECT SUM(time) as total FROM records
                             WHERE tkt_num = ?
                             AND created_on >= ?
                             AND created_on <= ?")
            _ (stmt:bind_values ticket.id start-date end-date)]
        (each [row (stmt:nrows)]
          (when (. row "total")
            (tset hours ticket.id
                  {:title ticket.title
                   :minutes (. row "total")})))))
    (db:close)
    hours))

(fn inv/validate-project-requirements [project]
  "Validate all requirements are met before generating a project invoice"
  (let [requirements []]

    ;; Check project has a client
    (when (or (= project.client nil) (= project.client ""))
      (table.insert requirements "Project must have a client assigned"))

    ;; Check project has a company
    (when (or (= project.company nil) (= project.company ""))
      (table.insert requirements "Project must have a company assigned"))

    ;; Check company exists and has address
    (when project.company
      (let [company (comp/info project.company)]
        (when (not company)
          (table.insert requirements (.. "Company '" project.company "' not found in database")))
        (when (and company (or (= (. company "address") nil) (= (. company "address") "")))
          (table.insert requirements (.. "Company '" project.company "' needs an address")))))

    ;; Check project has tickets
    (when (= (length project.tickets) 0)
      (table.insert requirements "Project has no tickets assigned"))

    requirements))

(fn proj/set-status [pid status]
  "Update project status"
  ;; pid is not passed as { project = pid } here so this fails.
  ;; FIXME another edge case util/err
  ;;(util/err "project" pid "A project ID is required.")
  (let [db (sql.open conf.tkt_loc)
        valid-statuses {"Active" true "On Hold" true "Completed" true}]
    (when (not (. valid-statuses status))
      (io.stderr:write "Invalid status. Must be 'Active', 'On Hold', or 'Completed'.\n")
      (os.exit))

    (let [stmt (db:prepare "UPDATE projects SET status = ? WHERE id = ?")]
      (stmt:bind_values status pid)
      (stmt:step)
      (stmt:finalize))
    (db:close)))

(fn proj/complete [args]
  "Mark a project as completed"
  (util/err "project" args "A project ID is required to complete a project.")
  (proj/set-status args.project "Completed")
  (print (.. "P" args.project " {:Status Completed}")))

(fn proj/hold [args]
  "Put a project on hold"
  (util/err "project" args "A project ID is required to put a project on hold.")
  (proj/set-status args.project "On Hold")
  (print (.. "P" args.project " {:Status On Hold}")))

(fn proj/reactivate [args]
  "Reactivate a project from hold or completed state"
  (util/err "project" args "A project ID is required to reactivate a project.")
  (proj/set-status args.project "Active")
  (print (.. "P" args.project " {:Status Active}")))

(fn proj/count [db ?filter]
  "Return count of projects by status"
  (let [sql "SELECT count(id) FROM projects "
        count {"active" 0 "on_hold" 0 "completed" 0 "total" 0}
        query (fn [sql targ]
                (each [row (db:nrows sql)]
                  (tset count targ (. row "count(id)"))))]

    (if (or (= ?filter nil) (= ?filter "Total"))
        (do
          (query (.. sql "WHERE status = 'Active'") "active")
          (query (.. sql "WHERE status = 'On Hold'") "on_hold")
          (query (.. sql "WHERE status = 'Completed'") "completed")
          (tset count "total" (+ count.active count.on_hold count.completed)))
        (= ?filter "Completed")
        (query (.. sql "WHERE status = 'Completed'") "completed")
        (= ?filter "Active")
        (query (.. sql "WHERE status = 'Active'") "active")
        (= ?filter "On Hold")
        (query (.. sql "WHERE status = 'On Hold'") "on_hold"))
    count))

(fn proj/format-project-line [row]
  "Format a project line for display"
  (.. "P" (. row "id") ") "
      "[" (or (. row "company") "---") "] - "
      (. row "name")))

(fn proj/list [status]
  "List projects by status"
  (let [db (sql.open conf.tkt_loc)
        stmt (db:prepare "SELECT id, name, company
                         FROM projects
                         WHERE status = ?
                         ORDER BY id DESC")]
    (stmt:bind_values status)
    (print (.. status " Projects:"))
    (var found false)
    (each [row (stmt:nrows)]
      (set found true)
      (print (proj/format-project-line row)))
    (when (not found)
      (print "  No projects found"))
    (stmt:finalize)
    (db:close)))

(fn proj/list-by-client [client]
  "List all projects for a client"
  (let [db (sql.open conf.tkt_loc)
        stmt (db:prepare "SELECT id, name, status, company
                         FROM projects
                         WHERE client = ?
                         ORDER BY status, id DESC")]
    (stmt:bind_values client)
    (print (.. "Projects for " client ":"))
    (var found false)
    (each [row (stmt:nrows)]
      (set found true)
      (print (.. (proj/format-project-line row)
                 " [" (. row "status") "]")))
    (when (not found)
      (print "  No projects found"))
    (stmt:finalize)
    (db:close)))

(fn proj/info [pid]
  "Get project information including associated tickets"
  (let [db (sql.open conf.tkt_loc)]
    (var project nil)
    ;; Get project details
    (each [row (db:nrows (.. "SELECT * FROM projects WHERE id = " pid))]
      (set project {:id pid
                    :name (. row "name")
                    :description (or (. row "description") "")
                    :status (. row "status")
                    :client (or (. row "client") "")
                    :company (or (. row "company") "")
                    :tickets []
                    :total_time 0}))

    ;; If project exists, get associated tickets
    (when project
      ;; Get associated tickets and sum time
      (each [row (db:nrows (.. "SELECT t.* FROM tickets t
                               JOIN project_tickets pt ON t.id = pt.ticket_id
                               WHERE pt.project_id = " pid))]
        (table.insert (. project :tickets)
                      {:id (. row "id")
                       :title (. row "title")
                       :status (. row "status")
                       :time (. row "time")})
        (tset project :total_time (+ (. project :total_time) (. row "time")))))

    (db:close)
    (or project
        {:id pid
         :name "Not Found"
         :description ""
         :status "Unknown"
         :client ""
         :company ""
         :tickets []
         :total_time 0})))

(fn proj/status [args]
  "Display project status"
  (let [project (proj/info args.project)]
    (if (= project.name "Not Found")
        (print (.. "Project P" args.project " not found"))
        (print (.. "P" args.project ") [" (or project.company "---") "] - " project.name "\n"
                   "Status: " project.status "\n"
                   "Description: " (or project.description "") "\n"
                   "Client: " (or project.client "") "\n"
                   "Total Time: " project.total_time "min\n"
                   "\nTickets:"
                   (if (= (length project.tickets) 0)
                       "\n  No tickets assigned"
                       ""))))
    (each [_ tkt (ipairs project.tickets)]
      (print (.. "  T" tkt.id ": " tkt.title " (" tkt.status ")")))))

(fn proj/report [args]
  "Generate project status report"
  (let [db (sql.open conf.tkt_loc)
        count (proj/count db)]
    (print (.. "Active: " count.active " | On Hold: " count.on_hold
               " | Completed: " count.completed))

    (if args.client
        (proj/list-by-client args.client)
        (if (or (= args.filter nil) (= (string.upper args.filter) "ACTIVE"))
            (proj/list "Active")
            (= (string.upper args.filter) "ON HOLD")
            (proj/list "On Hold")
            (= (string.upper args.filter) "COMPLETED")
            (proj/list "Completed")
            (= (string.upper args.filter) "ALL")
            (do
              (proj/list "Active")
              (proj/list "On Hold")
              (proj/list "Completed"))))
    (db:close)))

(fn proj/invoice [args]
  "Generate an invoice for a project, showing hours per ticket"
  (util/err "project" args "A project ID is required to generate an invoice.")

  (let [db (sql.open conf.tkt_loc)
        project (proj/info args.project)]

    ;; Validate requirements
    (let [requirements (inv/validate-project-requirements project)]
      (when (> (length requirements) 0)
        (io.stderr:write "Cannot generate project invoice. The following requirements must be met:\n")
        (each [_ req (ipairs requirements)]
          (io.stderr:write (.. "- " req "\n")))
        (os.exit)))

    (let [start-date (or args.start (os.date "%Y-%m-%d" 0))
          end-date (or args.end (os.date "%Y-%m-%d"))
          tax-rate (or args.tax_rate invoice-defaults.tax_rate)
          payment-terms (or args.payment_terms invoice-defaults.payment_terms)

          ;; Get project info and records
          companies (inv/get-abbreviations db)
          all-records []

          ;; Collect records from all project tickets
          _ (each [_ ticket (ipairs project.tickets)]
              (let [tkt-records (inv/get-records ticket.id start-date end-date)]
                (each [_ record (ipairs tkt-records)]
                  (table.insert all-records record))))

          ;; Group records by company abbreviation and tag
          groups (inv/group-by-abbrv all-records companies)
          line-items (inv/format-line-items groups (tonumber conf.rate))

          ;; Calculate totals
          total-minutes (accumulate [total 0
                                     _ record (ipairs all-records)]
                          (+ total record.time))
          total-amount (* (/ total-minutes 60) (tonumber conf.rate))
          tax-amount (* total-amount (/ tax-rate 100))
          final-amount (+ total-amount tax-amount)

          ;; Get user and company info
          user (clie/info conf.user conf.company)
          user_company (comp/info (. user "company"))
          client (clie/info project.client project.company)
          company (comp/info project.company)
          address (util/split (. company "address") ",")

          ;; Generate invoice number and dates
          number (inv/generate-number (.. "P" args.project) nil)
          issued-date (os.date "%x")
          due-date (os.date "%x" (+ (os.time) (* payment-terms 24 60 60)))
          period (.. start-date "-" end-date)

          ;; Format template variables
          template-vars (let [vars {:full_name (. user "name")
                                    :address (. user_company "address")
                                    :cell_number (. user "phone")
                                    :company (. client "company")
                                    :company_addr (.. (. address 1) "," (. address 2))
                                    :company_zip (string.gsub (. address 3) "^ " "")
                                    :client_email (. client "email")
                                    :client_cellphone (. client "phone")
                                    :invoice_number number
                                    :issued_date issued-date
                                    :due_date due-date
                                    :period period
                                    :line_items line-items
                                    :tax_rate tax-rate
                                    :total_cost (inv/format-currency total-amount)
                                    :tax_amount (inv/format-currency tax-amount)
                                    :final_amount (inv/format-currency final-amount)}]
                          (collect [k v (pairs vars)] k v))

          ;; Add template to vars and generate content
          _ (tset template-vars 1 (. invoice_templates (or args.template :default)))
          content (util/template template-vars)

          ;; Store invoice
          invoice {:number number
                   :tkt_num args.project
                   :client project.client
                   :content content
                   :start_date start-date
                   :end_date end-date
                   :tax_rate tax-rate
                   :payment_terms payment-terms
                   :final_amount final-amount}
          id (if args.update
                 (do
                   (inv/update db args.update invoice)
                   args.update)
                 (inv/store db invoice))]

      (print content)
      (db:close))))

;;===== status =====
;;(status/log 1 10)
(fn status/log [tkt limit]
  (each [k record (pairs (rec/info tkt limit))]
    (let
        [log (. record "log")
         user (. record "user")
         time (. record "time")]
      (print (..  k ") {" user ", " time "}\n")
             log))))

(fn status/tag-records [tnum tag]
  "Display records for a specific tag"
  (let [records (rec/get-tagged-records tnum tag)]
    (each [_ record (ipairs records)]
      (print (string.format "{%d, %dmin, %s}"
                            (. record "id")
                            (. record "time")
                            (. record "tag"))))))

(fn status/tag-summary [args]
  "Display time summary by tag for a ticket"
  (util/err "tkt" args "A tkt number is required to view tag summary.")
  (let [summary (rec/summary-by-tag args.tkt)]
    (print "Time by tag:")
    (each [tag data (pairs summary)]
      (print (.. "  " tag ": " data.time "min across " data.count " records")))))

;;(status/tkt {"tkt" 1 "verbose" true "records" 5})
(fn status/tkt [args]
  "Pretty print ticket information with optional tag filtering"
  (util/err "tkt" args "A tkt number is required to view info about a tkt.")
  (let [tkt (tkts/info args.tkt)]
    ;; Always show basic ticket info
    (print (.. (.. "Issue: " (or tkt.title "") "\n")
               (.. "Status: " tkt.status " | Owner: " (or tkt.owner "") "\n")
               (.. "Client: " (or tkt.client "") " | Proj: " (or tkt.project "") "\n")
               (.. "Desc: \n" (or tkt.desc "") "\n")
               (.. "Time: " tkt.time "min ($" tkt.rate ")")))

    (if args.tag
        ;; When filtering by tag, show tag-specific view
        (do
          (let [tag-time (rec/sum-tagged-time args.tkt args.tag)]
            (print (.. "\nLog: " tag-time "min tagged '" args.tag "'"))
            (status/tag-records args.tkt args.tag)))

        ;; Otherwise, show normal view based on verbose/records flags
        (do
          (if (and (= args.verbose nil) (= args.records nil))
              (do
                (print "Log: " tkt.log)
                (status/log args.tkt 1))
              (~= args.records nil)
              (do
                (status/log args.tkt args.records))
              (= args.verbose true)
              (do
                (status/log args.tkt)))))))

;;(status/comp {"name" "LambdaCreate"})
(fn status/comp [args]
  "Pretty print company information"
  (util/err "name" args "A company name is required to view info about a company.")
  (let
      [comp (comp/info args.name)]
    (if (= comp nil)
        (do
          (io.stderr:write "No company matching " args.name " found.\n")
          (os.exit))
        (do
          (print (.. (.. "Name: " comp.name "\n")
                     (.. "Abbrv: " comp.abbrv "\n")
                     (.. "Address: " comp.address "\n")
                     (.. "Comp ID: " comp.id "\n")
                     (.. "Create On: " comp.created_on)))))))

;;(status/clie {"name" "Will Sinatra" "company" "LambdaCreate"})
(fn status/clie [args]
  "Pretty print client information"
  (util/err "name" args "A client name is required to view information on a client.")
  (util/err "company" args "The company associated with the client is required to view information on a client.")
  (let
      [client (clie/info args.name args.company)]
    (if (= client nil)
        (do
          (io.stderr:write "No client found matching " args.name " in " args.company ".\n")
          (os.exit))
        (do
          (print (.. (.. "Name: " client.name "\n")
                     (.. "Company: " client.company "\n")
                     (.. "Role: " client.role "\n")
                     (.. "Phone: " client.phone "\n")
                     (.. "Email: " client.email "\n")
                     (.. "Client ID: " client.id "\n")
                     (.. "Created On: " client.created_on)))))))

;;(status/report {"filter" "all"})
(fn status/report [args]
  (let
      [db (sql.open conf.tkt_loc)
       count (tkts/count db)]
    (print (.. "Open: " count.open " | Closed: " count.closed))
    (if (or (= args.filter nil) (= (string.upper args.filter) "OPEN"))
        (tkts/list "Open")
        (= (string.upper args.filter) "CLOSED")
        (tkts/list "Closed")
        (= (string.upper args.filter) "ALL")
        (do
          (tkts/list "Open")
          (tkts/list "Closed")))))

;;(status/hours {"tkt" 1 "filter" "daily"})
;;filters: daily, weekly, monthly
;;((fn status/hours [args]
;;  "Generate hour breakdown report for a tkt"
;;  (util/err "tkt" args "A tkt is required to generate a report.")
;;  (util/err "filter" args "A filter is required, try daily, weekly, or monthly.")
;;  (let
;;      [db (sql.open conf.tkt_loc)
;;       report {"total_time" 0
;;               "daily_average" 0
;;               "weekly_average" 0
;;               "monthly_average" 0
;;               "daily" {}     ;; day_# = {"total" x}
;;               "weekly" {}    ;; week_# = {"total" x "average" x "day_1" x "day_7" x}
;;               "monthly" {}}]  ;; month_# = {"total" x "average" x "week_1" x "week_4" x}
;;       (each [row (db:nrows (.. "SELECT time,created_on from records WHERE tkt_num = '" args.tkt "';"))])))

(fn status/invoice-history [args]
  "Show invoice history for a client or ticket"
  (let [invoices (if args.client
                     (inv/get-by-client args.client)
                     (inv/get-by-ticket args.tkt))]
    (print (.. "Invoice History for " (if args.client (.. "Client: " args.client) (.. "Ticket: " args.tkt))))
    (print "----------------------------------------")
    (var total 0)
    (each [_ inv (ipairs invoices)]
      (print (string.format "Invoice %d: $%d (%s to %s)"
                            inv.id
                            inv.amount
                            inv.start_date
                            inv.end_date))
      (set total (+ total inv.amount)))
    (print "----------------------------------------")
    (print (string.format "Total Invoiced: $%d" total))))

;;=====[doc]=====
;;(tkts/doc {"tkt" 1})
(fn tkts/doc [args]
  "Generate a markdown document of a tkt's history"
  (let
      [tkt (tkts/info args.tkt)]
    (print (.. "# T" args.tkt " " tkt.title))
    (print tkt.desc)
    (each [k record (pairs (rec/info args.tkt nil))]
      (let
          [date (. record "created_on")
           log (. record "log")]
        (print (..  "## " date "\n" log "\n"))))))

;;===== usage =====
(local help-text
       {:header (.. ";;===== tkts " ver.tkts " =====\n"
                    "Author: Will Sinatra <tkts@lambdacreate.com>\n"
                    "License: GPLv3\n")

        :overview "usage: tkts <namespace:command> [<flags>]\n
help usage:
  tkts help                    |  Show ticket operations (default)
  tkts help -s|--segment all   |  Show all available commands
  tkts help -s|--segment NAME  |  Show commands for specific segment
                               |  Available segments: ticket, project, company, client, invoice

tkts namespaces:
  ticket    |  Ticket record operations (default)
  project   |  Project management
  company   |  Company record operations
  client    |  Client record operations
  invoice   |  Invoice generation

examples:
  tkts help -s ticket          |  Show all ticket commands
  tkts help -s project         |  Show all project commands
  tkts help -s all             |  Show all available commands
  tkts help -s brief           |  Show quick reference\n"

        :sections
        {:brief
         {:title ";;===== Quick Reference ====="
          :desc "Most commonly used commands."
          :commands
          ["# Create and manage tickets"
           "tkts create -i|--issue <title> [-d|--description <text>]"
           "tkts close -t|--tkt <number>"
           "tkts reopen -t|--tkt <number>"
           ""
           "# Track time and add logs"
           "tkts log -t|--tkt <number> [-l|--log <text>] [--tag <tag>]"
           "tkts time -t|--tkt <number> [--tag <tag>]"
           ""
           "# View status"
           "tkts info -t|--tkt <number> [-v|--verbose] [--tag <tag>]"
           "tkts status [-f|--filter open|closed|all]"
           ""
           "# Project management"
           "tkts project:create -n|--name <name> -c|--company <company>"
           "tkts project:edit -p|--project <id>"
           "tkts project:add-tkt -p|--project <id> -t|--tkt <id>"
           "tkts project:status -p|--project <id>"]}

         :ticket
         {:title ";;===== Ticket Operations ====="
          :desc "Create and manage tickets, add comments, and track time."
          :commands
          ["create -i|--issue <title> [-d|--description <text>]     |  Create a new ticket"
           "close -t|--tkt <number>                                 |  Close an existing ticket"
           "reopen -t|--tkt <number>                                |  Reopen a closed ticket"
           "info -t|--tkt <number> [--tag <tag>] [-v|--verbose]     |  Show ticket details"
           "log -t|--tkt <number> [-l|--log <text>] [--tag <tag>]   |  Add a log entry"
           "edit -t|--tkt <number> [-l|--log <number>]              |  Edit ticket or log entry"
           "time -t|--tkt <number> [--tag <tag>]                    |  Record time on ticket"
           "status [-f|--filter open|closed|all] [-v|--verbose]     |  Show ticket status"]}

         :project
         {:title ";;===== Project Operations ====="
          :desc "Manage projects and associated tickets."
          :commands
          ["project:create -n|--name <name> -c|--company <company> [--client <client>] [-d|--description <text>]  |  Create project"
           "project:edit -p|--project <id>                                                                        |  Edit a project"
           "project:add-tkt -p|--project <id> -t|--tkt <id>                                                       |  Add ticket to project"
           "project:rm-tkt -p|--project <id> -t|--tkt <id>                                                        |  Remove ticket from project"
           "project:status -p|--project <id> [-v|--verbose]                                                       |  Show project status"
           "project:complete -p|--project <id>                                                                    |  Mark project as completed"
           "project:hold -p|--project <id>                                                                        |  Put project on hold"
           "project:reactivate -p|--project <id>                                                                  |  Reactivate project"
           "project:invoice -p|--project <id> [--start <date>] [--end <date>]                                     |  Generate project invoice"
           "project:report [-f|--filter active|on hold|completed|all] [-c|--client <name>]                        |  Show project report"]}

         :company
         {:title ";;===== Company Operations ====="
          :desc "Manage company records and information."
          :commands
          ["company:create -n|--name <name> --abv|--abbreviation <abbr> [-addr|--address <addr>]  |  Add company"
           "company:edit -n|--name <name>                                                         |  Edit company"
           "company:list                                                                          |  List companies"
           "company:info -n|--name <name>                                                         |  Show company details"]}

         :client
         {:title ";;===== Client Operations ====="
          :desc "Manage client records and relationships."
          :commands
          ["client:create -n|--name <name> -c|--company <company>       |  Add client"
           "client:edit -n|--name <name> -c|--company <company>         |  Edit client"
           "client:list                                                 |  List clients"
           "client:info -n|--name <name> -c|--company <company>         |  Show client details"]}

         :invoice
         {:title ";;===== Invoice Operations ====="
          :desc "Generate and manage invoices."
          :commands
          ["invoice:create -t|--tkt <number> [--start <date>] [--end <date>]   |  Generate invoice from ticket"
           "invoice:list -c|--client <name>                                    |  Show invoice history for client"
           "invoice:list -t|--tkt <number>                                     |  Show invoice history for ticket"]}}})

(fn format-section [section]
  (.. section.title "\n"
      section.desc "\n\n"
      (table.concat section.commands "\n") "\n"))

(fn usage/generate [segment]
  "Generate usage information for specified segment"
  (let [output [help-text.header]]
    (match segment
      nil (do
            (table.insert output help-text.overview))

      "overview" (do
                   (table.insert output help-text.overview))

      "all" (do
              (table.insert output help-text.overview)
              (each [_ section (pairs help-text.sections)]
                (table.insert output (format-section section))))

      "brief" (do
                (table.insert output (format-section (. help-text.sections :brief))))

      _ (if (. help-text.sections segment)
            (do
              (table.insert output help-text.overview)
              (table.insert output (format-section (. help-text.sections segment))))
            (do
              (table.insert output help-text.overview)
              (table.insert output "\nError: Unknown segment. Use one of: ticket, project, company, client, invoice, brief, all\n"))))

    (table.concat output "\n")))

;;(main/usage {"verbose" true})
(fn main/usage [args]
  (let [segment (if (. args "segment") (. args "segment") nil)]
    (print (usage/generate segment))
    (os.exit)))

;; === Common Options ===
(local common-opts
       {:ticket {:name "tkt" :short "t" :description "Ticket number"}
        :issue {:name "issue" :short "i" :description "Issue title"}
        :description {:name "description" :short "d" :description "Detailed description"}
        :log {:name "log" :short "l" :description "Log message"}
        :tag {:name "tag" :description "Record tag"}
        :verbose {:name "verbose" :short "v" :description "Show detailed output"}
        :records {:name "records" :short "r" :description "Number of records to show"}
        :filter {:name "filter" :short "f" :description "Filter type (open/closed/all)"}
        :name {:name "name" :short "n" :description "Name"}
        :company {:name "company" :short "c" :description "Company name"}
        :abbreviation {:name "abbreviation" :short "abv" :description "Company abbreviation"}
        :address {:name "address" :short "addr" :description "Company address"}})

;; Project-specific options
(local project-opts
       {:project {:name "project" :short "p" :description "Project ID"}
        :ticket {:name "ticket" :short "t" :description "Ticket ID"}
        :start {:name "start" :description "Start date (YYYY-MM-DD)"}
        :end {:name "end" :description "End date (YYYY-MM-DD)"}})

;; === Command Registration ===
(cmd/register-alias "ls" "list")
(cmd/register-alias "new" "create")
(cmd/register-alias "rm" "remove")
(cmd/register-alias "mv" "move")

;; === Ticket Commands ===
(cmd/register "create"
              [common-opts.issue
               common-opts.description]
              (fn [opts] (tkts/create opts)))

(cmd/register "close"
              [common-opts.ticket]
              (fn [opts] (tkts/close opts)))

(cmd/register "reopen"
              [common-opts.ticket]
              (fn [opts] (tkts/reopen opts)))

(cmd/register "info"
              [common-opts.ticket
               common-opts.verbose
               common-opts.records
               common-opts.tag]
              (fn [opts] (status/tkt opts)))

(cmd/register "log"
              [common-opts.ticket
               common-opts.log
               common-opts.tag]
              (fn [opts] (tkts/comment opts)))

(cmd/register "edit"
              [common-opts.ticket
               common-opts.log]
              (fn [opts] (tkts/edit opts)))

(cmd/register "time"
              [common-opts.ticket
               common-opts.tag]
              (fn [opts] (tkts/time opts)))

;; tkts status has always been the canonical way to access this info
(cmd/register "status"
              [common-opts.filter
               common-opts.verbose]
              (fn [opts] (status/report opts)))

;; but every other module has a :list function, so this exists for compat.
(cmd/register "list"
              [common-opts.filter
               common-opts.verbose]
              (fn [opts] (status/report opts)))

(cmd/register "invoice:prepare"
              [common-opts.ticket
               {:name "client" :description "Client name"}]
              (fn [opts] (tkts/prepare-for-invoice opts)))

;; === Project Commands ===
(cmd/register "project:create"
              [common-opts.name
               common-opts.company
               common-opts.description
               {:name "client" :description "Client name"}]
              (fn [opts] (proj/create opts)))

;; Register project edit command
(cmd/register "project:edit"
              [project-opts.project]
              (fn [opts] (proj/edit opts)))

(cmd/register "project:add-tkt"
              [project-opts.project
               project-opts.ticket]
              (fn [opts] (proj/add-ticket opts)))

(cmd/register "project:rm-tkt"
              [project-opts.project
               project-opts.ticket]
              (fn [opts] (proj/remove-ticket opts)))

(cmd/register "project:status"
              [project-opts.project
               common-opts.verbose]
              (fn [opts] (proj/status opts)))

(cmd/register "project:complete"
              [project-opts.project]
              (fn [opts] (proj/complete opts)))

(cmd/register "project:hold"
              [project-opts.project]
              (fn [opts] (proj/hold opts)))

(cmd/register "project:reactivate"
              [project-opts.project]
              (fn [opts] (proj/reactivate opts)))

(cmd/register "project:invoice"
              [project-opts.project
               project-opts.start
               project-opts.end
               {:name "update" :description "Invoice ID to update"}]
              (fn [opts] (proj/invoice opts)))

;; project status is more of a report generation thing
(cmd/register "project:report"
              [{:name "filter" :short "f" :description "Filter (active/on hold/completed/all)"}
               {:name "client" :short "c" :description "Filter by client name"}]
              (fn [opts] (proj/report opts)))

;; but every other module has a :list function, so we have this for compat
(cmd/register "project:list"
              [{:name "filter" :short "f" :description "Filter (active/on hold/completed/all)"}
               {:name "client" :short "c" :description "Filter by client name"}]
              (fn [opts] (proj/report opts)))

;; === Company Commands ===
(cmd/register "company:create"
              [common-opts.name
               common-opts.abbreviation
               common-opts.address]
              (fn [opts] (comp/add opts)))

(cmd/register "company:edit"
              [common-opts.name]
              (fn [opts] (comp/edit opts)))

(cmd/register "company:list"
              []
              (fn [_] (comp/list)))

(cmd/register "company:info"
              [common-opts.name]
              (fn [opts] (status/comp opts)))

;; === Client Commands ===
(cmd/register "client:create"
              [common-opts.name
               common-opts.company]
              (fn [opts] (clie/add opts nil)))

(cmd/register "client:edit"
              [common-opts.name
               common-opts.company]
              (fn [opts] (clie/edit opts)))

(cmd/register "client:list"
              []
              (fn [_] (clie/list)))

(cmd/register "client:info"
              [common-opts.name
               common-opts.company]
              (fn [opts] (status/clie opts)))

;; === Invoice Commands ===
(cmd/register "invoice:create"
              [common-opts.ticket
               {:name "start" :description "Start date (YYYY-MM-DD)"}
               {:name "end" :description "End date (YYYY-MM-DD)"}
               {:name "update" :description "Invoice ID to update"}]
              (fn [opts] (tkts/invoice opts)))

(cmd/register "invoice:list"
              [{:name "client" :short "c" :description "Client name"}
               common-opts.ticket]
              (fn [opts] (status/invoice-history opts)))

;; === Help Command ===
(cmd/register "help"
              [{:name "segment" :short "s"
                :description "Help segment (ticket/project/company/client/invoice/brief/all)"
                :default "overview"}]
              (fn [opts]
                (main/usage {:segment (or opts.segment "overview")})))

;; === Default Command ===
(cmd/register "default"
              []
              (fn [_] (status/report {})))

(fn cmd/default-handler []
  "Default command handler showing status report"
  (status/report {}))

(fn cmd/unknown-command [cmd-path]
  "Handler for unknown commands"
  (io.stderr:write (string.format "Unknown command: %s\n" (table.concat cmd-path ":")))
  (io.stderr:write "Run 'tkts help' for usage information\n")
  (os.exit 1))

;; === Experimental Commands ===
(cmd/register "mail"
              [common-opts.ticket]
              (fn [opts] (tkts/mail opts)))

(cmd/register "doc"
              [common-opts.ticket]
              (fn [opts] (tkts/doc opts)))

;; ===== main =====
(fn main [args]
  (util/conf)
  (let [(cmd cmd-path remaining) (cmd/get args)]
    (if cmd
        (let [opts (cmd/parse-opts cmd.opts remaining)]
          (cmd.handler opts))
        (cmd/unknown-command cmd-path))))

(main arg)
